@@ -1,38 +1,17 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtScxml module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include <QtTest>
+#include <QtTest/private/qpropertytesthelper_p.h>
 #include <QObject>
 #include <QXmlStreamReader>
 #include <QtScxml/qscxmlcompiler.h>
 #include <QtScxml/qscxmlstatemachine.h>
 #include <QtScxml/qscxmlinvokableservice.h>
 #include <QtScxml/private/qscxmlstatemachine_p.h>
+#include <QtScxml/QScxmlNullDataModel>
+
+#include "topmachine.h"
 
 enum { SpyWaitTime = 8000 };
 
@@ -57,6 +36,8 @@ private Q_SLOTS:
 
     void multipleInvokableServices(); // QTBUG-61484
     void logWithoutExpr();
+
+    void bindings();
 };
 
 void tst_StateMachine::stateNames_data()
@@ -92,7 +73,7 @@ void tst_StateMachine::stateNames()
 
     QScopedPointer<QScxmlStateMachine> stateMachine(QScxmlStateMachine::fromFile(scxmlFileName));
     QVERIFY(!stateMachine.isNull());
-    QCOMPARE(stateMachine->parseErrors().count(), 0);
+    QCOMPARE(stateMachine->parseErrors().size(), 0);
 
     QCOMPARE(stateMachine->stateNames(compressed), expectedStates);
 }
@@ -383,7 +364,7 @@ void tst_StateMachine::doneDotStateEvent()
 
     stateMachine->start();
     finishedSpy.wait(5000);
-    QCOMPARE(finishedSpy.count(), 1);
+    QCOMPARE(finishedSpy.size(), 1);
     QCOMPARE(stateMachine->activeStateNames(true).size(), 1);
     QVERIFY(stateMachine->activeStateNames(true).contains(QLatin1String("success")));
 }
@@ -400,12 +381,12 @@ void tst_StateMachine::running()
 
     stateMachine->start();
 
-    QCOMPARE(runningChangedSpy.count(), 1);
+    QCOMPARE(runningChangedSpy.size(), 1);
     QCOMPARE(stateMachine->isRunning(), true);
 
     stateMachine->stop();
 
-    QCOMPARE(runningChangedSpy.count(), 2);
+    QCOMPARE(runningChangedSpy.size(), 2);
     QCOMPARE(stateMachine->isRunning(), false);
 }
 
@@ -419,8 +400,8 @@ void tst_StateMachine::invokeStateMachine()
     QCOMPARE(stateMachine->isRunning(), true);
     QTRY_VERIFY(stateMachine->activeStateNames().contains(QString("anyplace")));
 
-    QVector<QScxmlInvokableService *> services = stateMachine->invokedServices();
-    QCOMPARE(services.length(), 1);
+    QList<QScxmlInvokableService *> services = stateMachine->invokedServices();
+    QCOMPARE(services.size(), 1);
     QVariant subMachineVariant = services[0]->property("stateMachine");
     QVERIFY(subMachineVariant.isValid());
     QScxmlStateMachine *subMachine = qvariant_cast<QScxmlStateMachine *>(subMachineVariant);
@@ -439,7 +420,7 @@ void tst_StateMachine::multipleInvokableServices()
     QCOMPARE(stateMachine->isRunning(), true);
 
     finishedSpy.wait(5000);
-    QCOMPARE(finishedSpy.count(), 1);
+    QCOMPARE(finishedSpy.size(), 1);
     QCOMPARE(stateMachine->activeStateNames(true).size(), 1);
     QVERIFY(stateMachine->activeStateNames(true).contains(QLatin1String("success")));
 }
@@ -452,11 +433,109 @@ void tst_StateMachine::logWithoutExpr()
     QTest::ignoreMessage(QtDebugMsg, "\"Hi2\" : \"\"");
     stateMachine->start();
     QSignalSpy logSpy(stateMachine.data(), SIGNAL(log(QString,QString)));
-    QTRY_COMPARE(logSpy.count(), 1);
+    QTRY_COMPARE(logSpy.size(), 1);
+}
+
+void tst_StateMachine::bindings()
+{
+    // -- QScxmlStateMachine::initialized
+    std::unique_ptr<QScxmlStateMachine> stateMachine1(
+                QScxmlStateMachine::fromFile(QString(":/tst_statemachine/invoke.scxml")));
+    QVERIFY(stateMachine1.get());
+    QTestPrivate::testReadOnlyPropertyBasics<QScxmlStateMachine, bool>(
+                *stateMachine1, false, true, "initialized", [&](){ stateMachine1.get()->start(); });
+    if (QTest::currentTestFailed()) {
+        qWarning() << "QScxmlStateMachine::initialized bindable test failed.";
+        return;
+    }
+
+    // -- QScxmlStateMachine::initialValues
+    QVariantMap map1{{"map", 1}};
+    QVariantMap map2{{"map", 2}};
+    QTestPrivate::testReadWritePropertyBasics<QScxmlStateMachine, QVariantMap>(
+                *stateMachine1, map1, map2, "initialValues");
+    if (QTest::currentTestFailed()) {
+        qWarning() << "QScxmlStateMachine::initialValues bindable test failed.";
+        return;
+    }
+
+    // -- QScxmlStateMachine::loader
+    class MockLoader: public QScxmlCompiler::Loader
+    {
+    public:
+        QByteArray load(const QString&, const QString&, QStringList*) override { return QByteArray(); }
+    };
+    MockLoader loader1;
+    MockLoader loader2;
+    QTestPrivate::testReadWritePropertyBasics<QScxmlStateMachine, QScxmlCompiler::Loader*>(
+                *stateMachine1, &loader1, &loader2, "loader");
+    if (QTest::currentTestFailed()) {
+        qWarning() << "QScxmlStateMachine::loader bindable test failed.";
+        return;
+    }
+
+    // -- QScxmlStateMachine::dataModel
+    // Use non-existent file below, as valid file would initialize the model
+    std::unique_ptr<QScxmlStateMachine> stateMachine2(
+                QScxmlStateMachine::fromFile(QString("not_a_real_file")));
+    std::unique_ptr<QScxmlStateMachine> stateMachine3(
+                QScxmlStateMachine::fromFile(QString("not_a_real_file")));
+    QScxmlNullDataModel model1;
+    // data can only change once
+    QTestPrivate::testWriteOncePropertyBasics<QScxmlStateMachine, QScxmlDataModel*>(
+                *stateMachine2, nullptr, &model1, "dataModel");
+    if (QTest::currentTestFailed()) {
+        qWarning() << "QScxmlStateMachine::dataModel bindable test failed.";
+        return;
+    }
+
+    // -- QScxmlStateMachine::tableData
+    // Use the statemachine to generate the tabledDatas for testing
+    std::unique_ptr<QScxmlStateMachine> stateMachine4(
+                QScxmlStateMachine::fromFile(QString(":/tst_statemachine/invoke.scxml")));
+    QTestPrivate::testReadWritePropertyBasics<QScxmlStateMachine, QScxmlTableData*>(
+                *stateMachine2, stateMachine1.get()->tableData(), stateMachine4.get()->tableData(), "tableData");
+    if (QTest::currentTestFailed()) {
+        qWarning() << "QScxmlStateMachine::tableData bindable test failed.";
+        return;
+    }
+
+    // -- QScxmlStateMachine::invokedServices
+    // Test executes statemachine and observes as the invoked services change
+    TopMachine topSm;
+    QSignalSpy invokedSpy(&topSm, SIGNAL(invokedServicesChanged(const QList<QScxmlInvokableService *>)));
+    QCOMPARE(topSm.invokedServices().size(), 0);
+    // at some point during the topSm execution there are 3 invoked services
+    topSm.start();
+    QTRY_COMPARE(topSm.invokedServices().size(), 3);
+    QCOMPARE(invokedSpy.size(), 1);
+    // after completion invoked services drop back to 0
+    QTRY_COMPARE(topSm.isRunning(), false);
+    QCOMPARE(topSm.invokedServices().size(), 0);
+    QCOMPARE(invokedSpy.size(), 2);
+    // bind *to* the invokedservices property and check that we observe same changes
+    // during the topSm execution
+    QProperty<qsizetype> invokedServicesObserver;
+    invokedServicesObserver.setBinding([&](){ return topSm.invokedServices().size(); });
+    QCOMPARE(invokedServicesObserver, 0);
+    topSm.start();
+    QTRY_COMPARE(invokedServicesObserver, 3);
+    QCOMPARE(topSm.invokedServices().size(), 3);
+    QCOMPARE(invokedSpy.size(), 3);
+
+    // -- QScxmlDataModel::stateMachine
+    QScxmlNullDataModel dataModel1;
+    std::unique_ptr<QScxmlStateMachine> stateMachine5(
+                QScxmlStateMachine::fromFile(QString("not_a_real_file")));
+    // data can only change once
+    QTestPrivate::testWriteOncePropertyBasics<QScxmlDataModel, QScxmlStateMachine*>(
+                dataModel1, nullptr, stateMachine5.get(), "stateMachine");
+    if (QTest::currentTestFailed()) {
+        qWarning() << "QScxmlDataModel::stateMachine bindable test failed.";
+        return;
+    }
 }
 
 QTEST_MAIN(tst_StateMachine)
 
 #include "tst_statemachine.moc"
-
-
